/**
 * @author zhixin wen <wenzhixin2010@gmail.com>
 * version: 1.14.2
 * https://github.com/wenzhixin/bootstrap-table/
 */

import Constants from './constants/index.js'
import Utils from './utils/index.js'
import VirtualScroll from './virtual-scroll/index.js'
import {isNumeric, isEmptyObject} from './utils/types.js'
import Sort from './utils/sort.js'
import Polyfill from './dom/polyfill.js'
import {createElem, removeElem, addClass, removeClass, show, hide, find, appendTo, outerHeight} from './dom/dom.js'
import {addEvent, removeEvent} from './events/event.js'

class BootstrapTable {
  constructor (el, options) {
    Polyfill()
    this.options = options
    this.$el = $(el)
    this.$el_ = this.$el.clone()
    this.timeoutId_ = 0
    this.timeoutFooter_ = 0

    this.init()
  }

  init () {
    this.initConstants()
    this.initLocale()
    this.initContainer()
    this.initTable()
    this.initHeader()
    this.initData()
    this.initHiddenRows()
    this.initToolbar()
    this.initPagination()
    this.initBody()
    this.initSearchText()
    this.initServer()
  }

  initConstants () {
    const o = this.options
    this.constants = Constants.CONSTANTS
    this.constants.theme = $.fn.bootstrapTable.theme

    const buttonsPrefix = o.buttonsPrefix ? `${o.buttonsPrefix}-` : ''
    this.constants.buttonsClass = [
      o.buttonsPrefix,
      buttonsPrefix + o.buttonsClass,
      Utils.sprintf(`${buttonsPrefix}%s`, o.iconSize)
    ].join(' ').trim()
  }

  initLocale () {
    if (this.options.locale) {
      const locales = $.fn.bootstrapTable.locales
      const parts = this.options.locale.split(/-|_/)

      parts[0] = parts[0].toLowerCase()
      if (parts[1]) {
        parts[1] = parts[1].toUpperCase()
      }

      if (locales[this.options.locale]) {
        Utils.extend(this.options, locales[this.options.locale])
      } else if (locales[parts.join('-')]) {
        Utils.extend(this.options, locales[parts.join('-')])
      } else if (locales[parts[0]]) {
        Utils.extend(this.options, locales[parts[0]])
      }
    }
  }

  initContainer () {
    const topPagination = ['top', 'both'].includes(this.options.paginationVAlign)
      ? '<div class="fixed-table-pagination clearfix"></div>' : ''
    const bottomPagination = ['bottom', 'both'].includes(this.options.paginationVAlign)
      ? '<div class="fixed-table-pagination"></div>' : ''

    this.$container = createElem(`<div class="bootstrap-table ${this.constants.theme}">
      <div class="fixed-table-toolbar"></div>
      ${topPagination}
      <div class="fixed-table-container">
      <div class="fixed-table-header"><table></table></div>
      <div class="fixed-table-body">
      <div class="fixed-table-loading">
      <span class="loading-wrap">
      <span class="loading-text">${this.options.formatLoadingMessage()}</span>
      <span class="animation-wrap"><span class="animation-dot"></span></span>
      </span>
      </div>
      </div>
      <div class="fixed-table-footer"><table><thead><tr></tr></thead></table></div>
      </div>
      ${bottomPagination}
      </div>
    `)

    // Still support to jQuery
    this.$container = $(this.$container)

    this.$container.insertAfter(this.$el)
    this.$tableContainer = this.$container.find('.fixed-table-container')
    this.$tableHeader = this.$container.find('.fixed-table-header')
    this.$tableBody = this.$container.find('.fixed-table-body')
    this.$tableLoading = this.$container.find('.fixed-table-loading')
    this.$tableFooter = this.$el.find('tfoot')
    // checking if custom table-toolbar exists or not
    if (this.options.buttonsToolbar) {
      this.$toolbar = find(document.getElementsByTagName('body')[0], this.options.buttonsToolbar)[0]
    } else {
      this.$toolbar = find(this.$container, '.fixed-table-toolbar')[0]
    }

    this.$pagination = find(this.$container, '.fixed-table-pagination')[0]

    this.$tableBody.append(this.$el)
    this.$container.after('<div class="clearfix"></div>')

    addClass(this.$el, this.options.classes)
    addClass(this.$tableLoading, this.options.classes)

    if (this.options.height) {
      addClass(this.$tableContainer, 'fixed-height')

      if (this.options.showFooter) {
        addClass(this.$tableContainer, 'has-footer')
      }

      if (this.options.classes.split(' ').includes('table-bordered')) {
        this.$tableBody.append('<div class="fixed-table-border"></div>')
        this.$tableBorder = this.$tableBody.find('.fixed-table-border')
        addClass(this.$tableLoading, 'fixed-table-border')
      }

      this.$tableFooter = this.$container.find('.fixed-table-footer')
    }
  }

  initTable () {
    const columns = []
    const data = []

    this.$header = find(this.$el, 'thead')
    if (!this.$header.length) {
      this.$header = createElem('thead', ['class', this.options.theadClasses])
      appendTo(this.$el, this.$header)
    } else if (this.options.theadClasses) {
      addClass(this.$header, this.options.theadClasses)
    }
    find(this.$header, 'tr').forEach((tr, i) => {
      const column = []

      find(tr, 'th').forEach((i, th) => {
        // #2014: getFieldIndex and elsewhere assume this is string, causes issues if not
        if (typeof th.getAttribute('data-field') !== 'undefined') {
          th.setAttribute('data-field', th.getAttribute('data-field').toString())
        }
        column.push(Utils.extend({}, {
          title: $(th).html(),
          'class': th.getAttribute('class'),
          titleTooltip: th.getAttribute('title'),
          rowspan: $(th).attr('rowspan') ? +$(th).attr('rowspan') : undefined,
          colspan: $(th).attr('colspan') ? +$(th).attr('colspan') : undefined
        }, $(th).data()))
      })
      columns.push(column)
    })

    if (!Array.isArray(this.options.columns[0])) {
      this.options.columns = [this.options.columns]
    }

    this.options.columns = Utils.extend(true, [], columns, this.options.columns)
    this.columns = []
    this.fieldsColumnsIndex = []

    Utils.setFieldIndex(this.options.columns)

    this.options.columns.forEach((columns, i) => {
      columns.forEach((_column, j) => {
        const column = Utils.extend({}, BootstrapTable.COLUMN_DEFAULTS, _column)

        if (typeof column.fieldIndex !== 'undefined') {
          this.columns[column.fieldIndex] = column
          this.fieldsColumnsIndex[column.field] = column.fieldIndex
        }

        this.options.columns[i][j] = column
      })
    })

    // if options.data is setting, do not process tbody and tfoot data
    if (!this.options.data.length) {
      this.options.data = Utils.trToData(this.columns, find(this.$el, 'tbody>tr'))
      if (data.length) {
        this.fromHtml = true
      }
    }

    this.footerData = Utils.trToData(this.columns, find(this.$el, 'tfoot>tr'))
    if (this.footerData.length) {
      find(this.$el, 'tfoot')[0].appendChild(createElem('tr'))
    }

    if (!this.options.showFooter || this.options.cardView) {
      hide(this.$tableFooter)
    } else {
      show(this.$tableFooter)
    }
  }

  initHeader () {
    const visibleColumns = {}
    const html = []

    this.header = {
      fields: [],
      styles: [],
      classes: [],
      formatters: [],
      detailFormatters: [],
      events: [],
      sorters: [],
      sortNames: [],
      cellStyles: [],
      searchables: []
    }

    this.options.columns.forEach((columns, i) => {
      html.push('<tr>')

      if (i === 0 && !this.options.cardView && this.options.detailView && this.options.detailViewIcon) {
        html.push(`<th class="detail" rowspan="${this.options.columns.length}">
          <div class="fht-cell"></div>
          </th>
        `)
      }

      columns.forEach((column, j) => {
        const class_ = Utils.sprintf(' class="%s"', column['class'])
        const unitWidth = column.widthUnit
        const width = Number.parseFloat(column.width)

        const halign = Utils.sprintf('text-align: %s; ', column.halign ? column.halign : column.align)
        const align = Utils.sprintf('text-align: %s; ', column.align)
        let style = Utils.sprintf('vertical-align: %s; ', column.valign)
        style += Utils.sprintf('width: %s; ', (column.checkbox || column.radio) && !width
          ? (!column.showSelectTitle ? '36px' : undefined)
          : (width ? width + unitWidth : undefined))

        if (typeof column.fieldIndex !== 'undefined') {
          this.header.fields[column.fieldIndex] = column.field
          this.header.styles[column.fieldIndex] = align + style
          this.header.classes[column.fieldIndex] = class_
          this.header.formatters[column.fieldIndex] = column.formatter
          this.header.detailFormatters[column.fieldIndex] = column.detailFormatter
          this.header.events[column.fieldIndex] = column.events
          this.header.sorters[column.fieldIndex] = column.sorter
          this.header.sortNames[column.fieldIndex] = column.sortName
          this.header.cellStyles[column.fieldIndex] = column.cellStyle
          this.header.searchables[column.fieldIndex] = column.searchable

          if (!column.visible) {
            return
          }

          if (this.options.cardView && (!column.cardVisible)) {
            return
          }

          visibleColumns[column.field] = column
        }

        html.push(`<th${Utils.sprintf(' title="%s"', column.titleTooltip)}`,
          column.checkbox || column.radio
            ? Utils.sprintf(' class="bs-checkbox %s"', column['class'] || '')
            : class_,
          Utils.sprintf(' style="%s"', halign + style),
          Utils.sprintf(' rowspan="%s"', column.rowspan),
          Utils.sprintf(' colspan="%s"', column.colspan),
          Utils.sprintf(' data-field="%s"', column.field),
          // If `column` is not the first element of `this.options.columns[0]`, then className 'data-not-first-th' should be added.
          j === 0 && i > 0 ? ' data-not-first-th' : '',
          '>')

        html.push(Utils.sprintf('<div class="th-inner %s">', this.options.sortable && column.sortable
          ? 'sortable both' : ''))

        let text = this.options.escape ? Utils.escapeHTML(column.title) : column.title

        const title = text
        if (column.checkbox) {
          text = ''
          if (!this.options.singleSelect && this.options.checkboxHeader) {
            text = '<label><input name="btSelectAll" type="checkbox" /><span></span></label>'
          }
          this.header.stateField = column.field
        }
        if (column.radio) {
          text = ''
          this.header.stateField = column.field
          this.options.singleSelect = true
        }
        if (!text && column.showSelectTitle) {
          text += title
        }

        html.push(text)
        html.push('</div>')
        html.push('<div class="fht-cell"></div>')
        html.push('</div>')
        html.push('</th>')
      })
      html.push('</tr>')
    })

    this.$header.innerHTML = ''
    this.$header.appendChild(createElem(html.join('')))
    find(this.$header, 'th[data-field]').forEach((el, i) => {
      $(el).data(visibleColumns[$(el).data('field')])
    })
    find(this.$container, '.th-inner').forEach((thInner, i) => {
      addEvent(thInner, 'click', (e) => {
        const $this = $(e.currentTarget)

        if (this.options.detailView && !$this.parent().hasClass('bs-checkbox')) {
          if ($this.closest('.bootstrap-table')[0] !== this.$container[0]) {
            return false
          }
        }

        if (this.options.sortable && $this.parent().data().sortable) {
          this.onSort(e)
        }
      })
    })

    find(this.$header, 'th').forEach((th, i) => {
      addEvent(th, 'keypress', (e) => {
        if (this.options.sortable && $(e.currentTarget).data().sortable) {
          const code = e.keyCode || e.which
          if (code === 13) { // Enter keycode
            this.onSort(e)
          }
        }
      })
    })

    const resizeEvent = `resize.bootstrap-table${this.$el.attr('id') || ''}`
    removeEvent(window, resizeEvent)
    if (!this.options.showHeader || this.options.cardView) {
      hide(this.$header)
      hide(this.$tableHeader)
      this.$tableLoading.css('top', 0)
    } else {
      show(this.$header)
      show(this.$tableHeader)
      this.$tableLoading.css('top', this.$header.offsetHeight + 1)
      // Assign the correct sortable arrow
      this.getCaret()
      addEvent(window, resizeEvent, (e) => this.resetWidth(e))
    }

    this.$selectAll = find(this.$header, '[name="btSelectAll"]')
    // Still support to jQuery
    this.$selectAll = $(this.$selectAll)

    removeEvent(this.$selectAll, 'click')
    addEvent(this.$selectAll, 'click', ({currentTarget}) => {
      const checked = $(currentTarget).prop('checked')
      this[checked ? 'checkAll' : 'uncheckAll']()
      this.updateSelected()
    })
  }

  initData (data, type) {
    if (type === 'append') {
      this.options.data = this.options.data.concat(data)
    } else if (type === 'prepend') {
      this.options.data = [].concat(data).concat(this.options.data)
    } else {
      this.options.data = data || this.options.data
    }

    this.data = this.options.data

    if (this.options.sidePagination === 'server') {
      return
    }
    this.initSort()
  }

  initSort () {
    let name = this.options.sortName
    const order = this.options.sortOrder === 'desc' ? -1 : 1
    const index = this.header.fields.indexOf(this.options.sortName)
    let timeoutId = 0

    if (index !== -1) {
      if (this.options.sortStable) {
        this.data.forEach((row, i) => {
          if (!row.hasOwnProperty('_position')) {
            row._position = i
          }
        })
      }

      if (this.options.customSort) {
        Utils.calculateObjectValue(this.options, this.options.customSort, [
          this.options.sortName,
          this.options.sortOrder,
          this.data
        ])
      } else {
        this.data.sort((a, b) => {
          if (this.header.sortNames[index]) {
            name = this.header.sortNames[index]
          }
          const aa = Utils.getItemField(a, name, this.options.escape)
          const bb = Utils.getItemField(b, name, this.options.escape)
          const value = Utils.calculateObjectValue(this.header, this.header.sorters[index], [aa, bb, a, b])

          if (value !== undefined) {
            if (this.options.sortStable && value === 0) {
              return order * (a._position - b._position)
            }
            return order * value
          }

          return Sort(aa, bb, order, this.options.sortStable)
        })
      }

      if (this.options.sortClass !== undefined) {
        clearTimeout(timeoutId)
        timeoutId = setTimeout(() => {
          removeClass(this.$el, this.options.sortClass)
          const index = find(this.$header, `[data-field="${this.options.sortName}"]`)[0].cellIndex
          addClass(find(this.$el, `tr td:nth-child(${index + 1})`), this.options.sortClass)
        }, 250)
      }
    }
  }

  onSort ({type, currentTarget}) {
    const $this = type === 'keypress' ? currentTarget : currentTarget.parentNode
    const $this_ = find(this.$header, 'th')[$this.cellIndex]

    removeElem(find(this.$header, 'span.order'))
    removeElem(find(this.$header_, 'span.order'))

    if (this.options.sortName === $this.getAttribute('data-field')) {
      this.options.sortOrder = this.options.sortOrder === 'asc' ? 'desc' : 'asc'
    } else {
      this.options.sortName = $this.getAttribute('data-field')
      if (this.options.rememberOrder) {
        this.options.sortOrder = $this.getAttribute('data-order') === 'asc' ? 'desc' : 'asc'
      } else {
        this.options.sortOrder = this.columns[this.fieldsColumnsIndex[$this.getAttribute('data-field')]].order
      }
    }
    this.trigger('sort', this.options.sortName, this.options.sortOrder)

    $this.setAttribute('data-order', this.options.sortOrder)
    if ($this_) {
      $this_.setAttribute('data-order', this.options.sortOrder)
    }

    // Assign the correct sortable arrow
    this.getCaret()

    if (this.options.sidePagination === 'server') {
      this.initServer(this.options.silentSort)
      return
    }

    this.initSort()
    this.initBody()
  }

  initToolbar () {
    const o = this.options
    let html = []
    let timeoutId = 0
    let $keepOpen
    let $search
    let switchableCount = 0

    const bsBars = find(this.$toolbar, '.bs-bars')
    if (bsBars && bsBars[0] && bsBars[0].childNodes.length) {
      appendTo(document.getElementsByTagName('body')[0], createElem(o.toolbar))
    }

    this.$toolbar.innerHTML = ''

    if (typeof o.toolbar === 'string' || typeof o.toolbar === 'object') {
      const newToolbar = createElem(Utils.sprintf('<div class="bs-bars %s-%s"></div>', this.constants.classes.pull, o.toolbarAlign))
      appendTo(this.$toolbar, newToolbar)
      this.$toolbar.appendChild(createElem(o.toolbar))
    }

    // showColumns, showToggle, showRefresh
    html = [`<div class="${[
      'columns',
      `columns-${o.buttonsAlign}`,
      this.constants.classes.buttonsGroup,
      `${this.constants.classes.pull}-${o.buttonsAlign}`
    ].join(' ')}">`]

    if (typeof o.icons === 'string') {
      o.icons = Utils.calculateObjectValue(null, o.icons)
    }

    if (o.showPaginationSwitch) {
      html.push(`<button class="${this.constants.buttonsClass}" type="button" name="paginationSwitch"
        aria-label="Pagination Switch" title="${o.formatPaginationSwitch()}">
        ${Utils.sprintf(this.constants.html.icon, o.iconsPrefix, o.icons.paginationSwitchDown)}
        </button>`)
    }

    if (o.showRefresh) {
      html.push(`<button class="${this.constants.buttonsClass}" type="button" name="refresh"
        aria-label="Refresh" title="${o.formatRefresh()}">
        ${Utils.sprintf(this.constants.html.icon, o.iconsPrefix, o.icons.refresh)}
        </button>`)
    }

    if (o.showToggle) {
      html.push(`<button class="${this.constants.buttonsClass}" type="button" name="toggle"
        aria-label="Toggle" title="${o.formatToggle()}">
        ${Utils.sprintf(this.constants.html.icon, o.iconsPrefix, o.icons.toggleOff)}
        </button>`)
    }

    if (o.showFullscreen) {
      html.push(`<button class="${this.constants.buttonsClass}" type="button" name="fullscreen"
        aria-label="Fullscreen" title="${o.formatFullscreen()}">
        ${Utils.sprintf(this.constants.html.icon, o.iconsPrefix, o.icons.fullscreen)}
        </button>`)
    }

    if (o.showColumns) {
      html.push(`<div class="keep-open ${this.constants.classes.buttonsDropdown}" title="${o.formatColumns()}">
        <button class="${this.constants.buttonsClass} dropdown-toggle" type="button" data-toggle="dropdown"
        aria-label="Columns" title="${o.formatColumns()}">
        ${Utils.sprintf(this.constants.html.icon, o.iconsPrefix, o.icons.columns)}
        ${this.constants.html.dropdownCaret}
        </button>
        ${this.constants.html.toobarDropdow[0]}`)

      this.columns.forEach((column, i) => {
        if (column.radio || column.checkbox) {
          return
        }

        if (o.cardView && !column.cardVisible) {
          return
        }

        const checked = column.visible ? ' checked="checked"' : ''

        if (column.switchable) {
          html.push(Utils.sprintf(this.constants.html.toobarDropdowItem,
            Utils.sprintf('<input type="checkbox" data-field="%s" value="%s"%s> <span>%s</span>',
              column.field, i, checked, column.title)))
          switchableCount++
        }
      })
      html.push(this.constants.html.toobarDropdow[1], '</div>')
    }

    html.push('</div>')

    // Fix #188: this.showToolbar is for extensions
    if (this.showToolbar || html.length > 2) {
      this.$toolbar.appendChild(createElem(html.join('')))
    }

    if (o.showPaginationSwitch) {
      const paginationSwitch = find(this.$toolbar, 'button[name="paginationSwitch"]')[0]
      removeEvent(paginationSwitch, 'click')
      addEvent(paginationSwitch, 'click', () => this.togglePagination())
    }

    if (o.showFullscreen) {
      const fullscreen = find(this.$toolbar, 'button[name="fullscreen"]')[0]
      removeEvent(fullscreen, 'click')
      addEvent(fullscreen, 'click', () => this.toggleFullscreen())
    }

    if (o.showRefresh) {
      const refresh = find(this.$toolbar, 'button[name="refresh"]')[0]
      removeEvent(refresh, 'click')
      addEvent(refresh, 'click', () => this.refresh())
    }

    if (o.showToggle) {
      const toggle = find(this.$toolbar, 'button[name="toggle"]')[0]
      removeEvent(toggle, 'click')
      addEvent(toggle, 'click', () => this.toggleView())
    }

    if (o.showColumns) {
      $keepOpen = find(this.$toolbar, '.keep-open')[0]

      if (switchableCount <= o.minimumCountColumns) {
        find($keepOpen, 'input').forEach((input, i) => {
          input.setAttribute('disabled', 'disabled')
        })
      }

      find($keepOpen, 'li, label').forEach((elem, i) => {
        removeEvent(elem, 'click')
        addEvent(elem, 'click', (e) => {
          e.stopImmediatePropagation()
        })
      })

      find($keepOpen, 'input').forEach((input, i) => {
        removeEvent(input, 'click')
        addEvent(input, 'click', ({currentTarget}) => {
          const $this = $(currentTarget)

          this.toggleColumn($this.val(), $this.prop('checked'), false)
          this.trigger('column-switch', $this.data('field'), $this.prop('checked'))
        })
      })
    }

    if (o.search) {
      html = []
      html.push(`<div class="${this.constants.classes.pull}-${o.searchAlign} search ${this.constants.classes.inputGroup}">
        <input class="${this.constants.classes.input}${Utils.sprintf(' input-%s', o.iconSize)}"
        type="text" placeholder="${o.formatSearch()}">
        </div>`)

      this.$toolbar.appendChild(createElem(html.join('')))
      $search = find(this.$toolbar, '.search input')[0]
      removeEvent($search, 'keyup drop blur')
      addEvent($search, 'keyup drop blur', (event) => {
        if (o.searchOnEnterKey && event.keyCode !== 13) {
          return
        }

        if ([37, 38, 39, 40].includes(event.keyCode)) {
          return
        }

        clearTimeout(timeoutId) // doesn't matter if it's 0
        timeoutId = setTimeout(() => {
          this.onSearch(event)
        }, o.searchTimeOut)
      })

      if (Utils.isIEBrowser()) {
        removeEvent($search, 'mouseup')
        addEvent($search, 'mouseup', (event) => {
          clearTimeout(timeoutId) // doesn't matter if it's 0
          timeoutId = setTimeout(() => {
            this.onSearch(event)
          }, o.searchTimeOut)
        })
      }
    }

    // Still support to jQuery
    this.$toolbar = $(this.$toolbar)
  }

  onSearch ({currentTarget, firedByInitSearchText}) {
    const target = $(currentTarget)
    const value = target.val()
    let text = ''
    if (Array.isArray(value)) {
      text = value[0].trim()
    } else {
      text = value.trim()
    }

    // trim search input
    if (this.options.trimOnSearch && target.val() !== text) {
      target.val(text)
    }

    if (text === this.searchText) {
      return
    }
    this.searchText = text
    this.options.searchText = text

    if (!firedByInitSearchText) {
      this.options.pageNumber = 1
    }
    this.initSearch()
    if (firedByInitSearchText) {
      if (this.options.sidePagination === 'client') {
        this.updatePagination()
      }
    } else {
      this.updatePagination()
    }
    this.trigger('search', text)
  }

  initSearch () {
    this.filterOptions = this.filterOptions || this.options.filterOptions
    if (this.options.sidePagination !== 'server') {
      if (this.options.customSearch) {
        this.data = Utils.calculateObjectValue(this.options, this.options.customSearch,
          [this.options.data, this.searchText])
        return
      }

      const s = this.searchText && (this.options.escape
        ? Utils.escapeHTML(this.searchText) : this.searchText).toLowerCase()
      const f = isEmptyObject(this.filterColumns) ? null : this.filterColumns

      // Check filter
      if (typeof this.filterOptions.filterAlgorithm === 'function') {
        this.data = this.options.data.filter((item, i) => this.filterOptions.filterAlgorithm.apply(null, [item, f]))
      } else if (typeof this.filterOptions.filterAlgorithm === 'string') {
        this.data = f ? this.options.data.filter((item, i) => {
          const filterAlgorithm = this.filterOptions.filterAlgorithm
          if (filterAlgorithm === 'and') {
            for (const key in f) {
              if (
                (Array.isArray(f[key]) &&
                  !f[key].includes(item[key])) ||
                (!Array.isArray(f[key]) &&
                  item[key] !== f[key])
              ) {
                return false
              }
            }
          } else if (filterAlgorithm === 'or') {
            let match = false
            for (const key in f) {
              if (
                (Array.isArray(f[key]) &&
                  f[key].includes(item[key])) ||
                (!Array.isArray(f[key]) &&
                  item[key] === f[key])
              ) {
                match = true
              }
            }

            return match
          }

          return true
        }) : this.options.data
      }

      this.data = s ? this.data.filter((item, i) => {
        for (let j = 0; j < this.header.fields.length; j++) {
          if (!this.header.searchables[j]) {
            continue
          }

          const key = isNumeric(this.header.fields[j]) ? parseInt(this.header.fields[j], 10) : this.header.fields[j]
          const column = this.columns[this.fieldsColumnsIndex[key]]
          let value

          if (typeof key === 'string') {
            value = item
            const props = key.split('.')
            for (let i = 0; i < props.length; i++) {
              if (value[props[i]] !== null) {
                value = value[props[i]]
              }
            }
          } else {
            value = item[key]
          }

          // Fix #142: respect searchForamtter boolean
          if (column && column.searchFormatter) {
            value = Utils.calculateObjectValue(column,
              this.header.formatters[j], [value, item, i, column.field], value)
          }

          if (typeof value === 'string' || typeof value === 'number') {
            if (this.options.strictSearch) {
              if ((`${value}`).toLowerCase() === s) {
                return true
              }
            } else {
              const largerSmallerEqualsRegex = /(?:(<=|=>|=<|>=|>|<)(?:\s+)?(\d+)?|(\d+)?(\s+)?(<=|=>|=<|>=|>|<))/gm
              const matches = largerSmallerEqualsRegex.exec(s)
              let comparisonCheck = false

              if (matches) {
                const operator = matches[1] || `${matches[5]}l`
                const comparisonValue = matches[2] || matches[3]
                const int = parseInt(value, 10)
                const comparisonInt = parseInt(comparisonValue, 10)

                switch (operator) {
                  case '>':
                  case '<l':
                    comparisonCheck = int > comparisonInt
                    break
                  case '<':
                  case '>l':
                    comparisonCheck = int < comparisonInt
                    break
                  case '<=':
                  case '=<':
                  case '>=l':
                  case '=>l':
                    comparisonCheck = int <= comparisonInt
                    break
                  case '>=':
                  case '=>':
                  case '<=l':
                  case '=<l':
                    comparisonCheck = int >= comparisonInt
                    break
                  default:
                    break
                }
              }

              if (comparisonCheck || (`${value}`).toLowerCase().includes(s)) {
                return true
              }
            }
          }
        }
        return false
      }) : this.data
    }
  }

  initPagination () {
    const o = this.options
    if (!o.pagination) {
      hide(this.$pagination)
      return
    }
    show(this.$pagination)

    const html = []
    let $allSelected = false
    let i
    let from
    let to
    let $pageList
    let $pre
    let $next
    let $number
    const data = this.getData()
    let pageList = o.pageList

    if (o.sidePagination !== 'server') {
      o.totalRows = data.length
    }

    this.totalPages = 0
    if (o.totalRows) {
      if (o.pageSize === o.formatAllRows()) {
        o.pageSize = o.totalRows
        $allSelected = true
      } else if (o.pageSize === o.totalRows) {
        // Fix #667 Table with pagination,
        // multiple pages and a search this matches to one page throws exception
        const pageLst = typeof o.pageList === 'string'
          ? o.pageList.replace('[', '').replace(']', '')
            .replace(/ /g, '').toLowerCase().split(',') : o.pageList
        if (pageLst.includes(o.formatAllRows().toLowerCase())) {
          $allSelected = true
        }
      }

      this.totalPages = ~~((o.totalRows - 1) / o.pageSize) + 1

      o.totalPages = this.totalPages
    }
    if (this.totalPages > 0 && o.pageNumber > this.totalPages) {
      o.pageNumber = this.totalPages
    }

    this.pageFrom = (o.pageNumber - 1) * o.pageSize + 1
    this.pageTo = o.pageNumber * o.pageSize
    if (this.pageTo > o.totalRows) {
      this.pageTo = o.totalRows
    }

    if (this.options.pagination && this.options.sidePagination !== 'server') {
      this.options.totalNotFiltered = this.options.data.length
    }

    if (!this.options.showExtendedPagination) {
      this.options.totalNotFiltered = undefined
    }

    const paginationInfo = o.onlyInfoPagination ?
      o.formatDetailPagination(o.totalRows) :
      o.formatShowingRows(this.pageFrom, this.pageTo, o.totalRows, o.totalNotFiltered)

    html.push(`<div class="${this.constants.classes.pull}-${o.paginationDetailHAlign} pagination-detail">
      <span class="pagination-info">
      ${paginationInfo}
      </span>`)

    if (!o.onlyInfoPagination) {
      html.push('<span class="page-list">')

      const pageNumber = [
        `<span class="${this.constants.classes.paginationDropdown}">
        <button class="${this.constants.buttonsClass} dropdown-toggle" type="button" data-toggle="dropdown">
        <span class="page-size">
        ${$allSelected ? o.formatAllRows() : o.pageSize}
        </span>
        ${this.constants.html.dropdownCaret}
        </button>
        ${this.constants.html.pageDropdown[0]}`]

      if (typeof o.pageList === 'string') {
        const list = o.pageList.replace('[', '').replace(']', '')
          .replace(/ /g, '').split(',')

        pageList = []
        for (const value of list) {
          pageList.push(
            (value.toLowerCase() === o.formatAllRows().toLowerCase() ||
              ['all', 'unlimited'].includes(value.toLowerCase()))
              ? o.formatAllRows() : +value)
        }
      }

      pageList.forEach((page, i) => {
        if (!o.smartDisplay || i === 0 || pageList[i - 1] < o.totalRows) {
          let active
          if ($allSelected) {
            active = page === o.formatAllRows() ? this.constants.classes.dropdownActive : ''
          } else {
            active = page === o.pageSize ? this.constants.classes.dropdownActive : ''
          }
          pageNumber.push(Utils.sprintf(this.constants.html.pageDropdownItem, active, page))
        }
      })
      pageNumber.push(`${this.constants.html.pageDropdown[1]}</span>`)

      html.push(o.formatRecordsPerPage(pageNumber.join('')))
      html.push('</span></div>')

      html.push(`<div class="${this.constants.classes.pull}-${o.paginationHAlign} pagination">`,
        Utils.sprintf(this.constants.html.pagination[0], Utils.sprintf(' pagination-%s', o.iconSize)),
        Utils.sprintf(this.constants.html.paginationItem, ' page-pre', o.paginationPreText))

      if (this.totalPages < o.paginationSuccessivelySize) {
        from = 1
        to = this.totalPages
      } else {
        from = o.pageNumber - o.paginationPagesBySide
        to = from + (o.paginationPagesBySide * 2)
      }

      if (o.pageNumber < (o.paginationSuccessivelySize - 1)) {
        to = o.paginationSuccessivelySize
      }

      if (o.paginationSuccessivelySize > this.totalPages - from) {
        from = from - (o.paginationSuccessivelySize - (this.totalPages - from)) + 1
      }

      if (from < 1) {
        from = 1
      }

      if (to > this.totalPages) {
        to = this.totalPages
      }

      const middleSize = Math.round(o.paginationPagesBySide / 2)
      const pageItem = (i, classes = '') => Utils.sprintf(this.constants.html.paginationItem,
        classes + (i === o.pageNumber ? ` ${this.constants.classes.paginationActive}` : ''), i)

      if (from > 1) {
        let max = o.paginationPagesBySide
        if (max >= from) max = from - 1
        for (i = 1; i <= max; i++) {
          html.push(pageItem(i))
        }
        if ((from - 1) === max + 1) {
          i = from - 1
          html.push(pageItem(i))
        } else {
          if ((from - 1) > max) {
            if (
              (from - o.paginationPagesBySide * 2) > o.paginationPagesBySide &&
              o.paginationUseIntermediate
            ) {
              i = Math.round(((from - middleSize) / 2) + middleSize)
              html.push(pageItem(i, ' page-intermediate'))
            } else {
              html.push(Utils.sprintf(this.constants.html.paginationItem,
                ' page-first-separator disabled', '...'))
            }
          }
        }
      }

      for (i = from; i <= to; i++) {
        html.push(pageItem(i))
      }

      if (this.totalPages > to) {
        let min = this.totalPages - (o.paginationPagesBySide - 1)
        if (to >= min) min = to + 1
        if ((to + 1) === min - 1) {
          i = to + 1
          html.push(pageItem(i))
        } else {
          if (min > (to + 1)) {
            if (
              (this.totalPages - to) > o.paginationPagesBySide * 2 &&
              o.paginationUseIntermediate
            ) {
              i = Math.round(((this.totalPages - middleSize - to) / 2) + to)
              html.push(pageItem(i, ' page-intermediate'))
            } else {
              html.push(Utils.sprintf(this.constants.html.paginationItem,
                ' page-last-separator disabled', '...'))
            }
          }
        }

        for (i = min; i <= this.totalPages; i++) {
          html.push(pageItem(i))
        }
      }

      html.push(Utils.sprintf(this.constants.html.paginationItem, ' page-next', o.paginationNextText))
      html.push(this.constants.html.pagination[1], '</div>')
    }
    this.$pagination.innerHTML = ''
    this.$pagination.appendChild(createElem(html.join('')))

    const dropupClass = ['bottom', 'both'].includes(o.paginationVAlign) ?
      ` ${this.constants.classes.dropup}` : ''
    addClass(find(this.$pagination, '.page-list > span')[0], dropupClass)

    if (!o.onlyInfoPagination) {
      $pageList = find(this.$pagination, '.page-list a')
      $pre = find(this.$pagination, '.page-pre')
      $next = find(this.$pagination, '.page-next')
      $number = find(this.$pagination, '.page-item:not(.page-next):not(.page-pre):not(.page-last-separator):not(.page-first-separator)')

      if (this.totalPages <= 1) {
        hide(find(this.$pagination, 'div.pagination'))
      }

      if (o.smartDisplay) {
        if (pageList.length < 2 || o.totalRows <= pageList[0]) {
          hide(find(this.$pagination, 'span.page-list'))
        }
      }

      // when data is empty, hide the pagination
      if (this.getData().length) {
        show(this.$pagination)
      } else {
        hide(this.$pagination)
      }

      if (!o.paginationLoop) {
        if (o.pageNumber === 1) {
          addClass($pre, 'disabled')
        }
        if (o.pageNumber === this.totalPages) {
          addClass($next, 'disabled')
        }
      }

      if ($allSelected) {
        o.pageSize = o.formatAllRows()
      }
      // removed the events for last and first, onPageNumber executeds the same logic
      removeEvent($pageList, 'click')
      addEvent($pageList, 'click', (e) => this.onPageListChange(e))

      removeEvent($pre, 'click')
      addEvent($pre, 'click', (e) => this.onPagePre(e))

      removeEvent($next, 'click')
      addEvent($next, 'click', (e) => this.onPageNext(e))

      removeEvent($number, 'click')
      addEvent($number, 'click', (e) => this.onPageNumber(e))
    }
  }

  updatePagination (event) {
    // Fix #171: IE disabled button can be clicked bug.
    if (event && $(event.currentTarget).hasClass('disabled')) {
      return
    }

    if (!this.options.maintainMetaData) {
      this.resetRows()
    }

    this.initPagination()
    if (this.options.sidePagination === 'server') {
      this.initServer()
    } else {
      this.initBody()
    }

    this.trigger('page-change', this.options.pageNumber, this.options.pageSize)
  }

  onPageListChange (event) {
    event.preventDefault()
    const $this = $(event.currentTarget)

    $this.parent().addClass(this.constants.classes.dropdownActive)
      .siblings().removeClass(this.constants.classes.dropdownActive)
    this.options.pageSize = $this.text().toUpperCase() === this.options.formatAllRows().toUpperCase()
      ? this.options.formatAllRows() : +$this.text()
    this.$toolbar.find('.page-size').text(this.options.pageSize)

    this.updatePagination(event)
    return false
  }

  onPagePre (event) {
    event.preventDefault()
    if ((this.options.pageNumber - 1) === 0) {
      this.options.pageNumber = this.options.totalPages
    } else {
      this.options.pageNumber--
    }
    this.updatePagination(event)
    return false
  }

  onPageNext (event) {
    event.preventDefault()
    if ((this.options.pageNumber + 1) > this.options.totalPages) {
      this.options.pageNumber = 1
    } else {
      this.options.pageNumber++
    }
    this.updatePagination(event)
    return false
  }

  onPageNumber (event) {
    event.preventDefault()
    if (this.options.pageNumber === +$(event.currentTarget).text()) {
      return
    }
    this.options.pageNumber = +$(event.currentTarget).text()
    this.updatePagination(event)
    return false
  }

  initRow (item, i) {
    const html = []
    let style = {}
    const csses = []
    let data_ = ''
    let attributes = {}
    const htmlAttributes = []

    if (Utils.findIndex(this.hiddenRows, item) > -1) {
      return
    }

    style = Utils.calculateObjectValue(this.options, this.options.rowStyle, [item, i], style)

    if (style && style.css) {
      for (const [key, value] of Object.entries(style.css)) {
        csses.push(`${key}: ${value}`)
      }
    }

    attributes = Utils.calculateObjectValue(this.options,
      this.options.rowAttributes, [item, i], attributes)

    if (attributes) {
      for (const [key, value] of Object.entries(attributes)) {
        htmlAttributes.push(`${key}="${Utils.escapeHTML(value)}"`)
      }
    }

    if (item._data && !isEmptyObject(item._data)) {
      for (const [k, v] of Object.entries(item._data)) {
        // ignore data-index
        if (k === 'index') {
          return
        }
        data_ += ` data-${k}='${typeof v === 'object' ? JSON.stringify(v) : v}'`
      }
    }

    html.push('<tr',
      Utils.sprintf(' %s', htmlAttributes.length ? htmlAttributes.join(' ') : undefined),
      Utils.sprintf(' id="%s"', Array.isArray(item) ? undefined : item._id),
      Utils.sprintf(' class="%s"', style.classes || (Array.isArray(item) ? undefined : item._class)),
      ` data-index="${i}"`,
      Utils.sprintf(' data-uniqueid="%s"', item[this.options.uniqueId]),
      Utils.sprintf(' data-has-detail-view="%s"', (!this.options.cardView && this.options.detailView && Utils.calculateObjectValue(null, this.options.detailFilter, [i, item])) ? 'true' : undefined),
      Utils.sprintf('%s', data_),
      '>'
    )

    if (this.options.cardView) {
      html.push(`<td colspan="${this.header.fields.length}"><div class="card-views">`)
    }

    if (!this.options.cardView && this.options.detailView && this.options.detailViewIcon) {
      html.push('<td>')

      if (Utils.calculateObjectValue(null, this.options.detailFilter, [i, item])) {
        html.push(`
          <a class="detail-icon" href="#">
          ${Utils.sprintf(this.constants.html.icon, this.options.iconsPrefix, this.options.icons.detailOpen)}
          </a>
        `)
      }

      html.push('</td>')
    }

    this.header.fields.forEach((field, j) => {
      let text = ''
      let value_ = Utils.getItemField(item, field, this.options.escape)
      let value = ''
      let type = ''
      let cellStyle = {}
      let id_ = ''
      let class_ = this.header.classes[j]
      let style_ = ''
      let data_ = ''
      let rowspan_ = ''
      let colspan_ = ''
      let title_ = ''
      const column = this.columns[j]

      if (this.fromHtml && typeof value_ === 'undefined') {
        if ((!column.checkbox) && (!column.radio)) {
          return
        }
      }

      if (!column.visible) {
        return
      }

      if (this.options.cardView && (!column.cardVisible)) {
        return
      }

      if (column.escape) {
        value_ = Utils.escapeHTML(value_)
      }

      if (csses.concat([this.header.styles[j]]).length) {
        style_ = ` style="${csses.concat([this.header.styles[j]]).join('; ')}"`
      }
      // handle td's id and class
      if (item[`_${field}_id`]) {
        id_ = Utils.sprintf(' id="%s"', item[`_${field}_id`])
      }
      if (item[`_${field}_class`]) {
        class_ = Utils.sprintf(' class="%s"', item[`_${field}_class`])
      }
      if (item[`_${field}_rowspan`]) {
        rowspan_ = Utils.sprintf(' rowspan="%s"', item[`_${field}_rowspan`])
      }
      if (item[`_${field}_colspan`]) {
        colspan_ = Utils.sprintf(' colspan="%s"', item[`_${field}_colspan`])
      }
      if (item[`_${field}_title`]) {
        title_ = Utils.sprintf(' title="%s"', item[`_${field}_title`])
      }
      cellStyle = Utils.calculateObjectValue(this.header,
        this.header.cellStyles[j], [value_, item, i, field], cellStyle)
      if (cellStyle.classes) {
        class_ = ` class="${cellStyle.classes}"`
      }
      if (cellStyle.css) {
        const csses_ = []
        for (const [key, value] of Object.entries(cellStyle.css)) {
          csses_.push(`${key}: ${value}`)
        }
        style_ = ` style="${csses_.concat(this.header.styles[j]).join('; ')}"`
      }

      value = Utils.calculateObjectValue(column,
        this.header.formatters[j], [value_, item, i, field], value_)

      if (item[`_${field}_data`] && !isEmptyObject(item[`_${field}_data`])) {
        for (const [k, v] of Object.entries(item[`_${field}_data`])) {
          // ignore data-index
          if (k === 'index') {
            return
          }
          data_ += ` data-${k}="${v}"`
        }
      }

      if (column.checkbox || column.radio) {
        type = column.checkbox ? 'checkbox' : type
        type = column.radio ? 'radio' : type

        const c = column['class'] || ''
        const isChecked = (value === true || value_ || (value && value.checked)) && value !== false
        const isDisabled = !column.checkboxEnabled || (value && value.disabled)

        text = [
          this.options.cardView
            ? `<div class="card-view ${c}">`
            : `<td class="bs-checkbox ${c}"${class_}${style_}>`,
          `<label>
            <input
            data-index="${i}"
            name="${this.options.selectItemName}"
            type="${type}"
            ${Utils.sprintf('value="%s"', item[this.options.idField])}
            ${Utils.sprintf('checked="%s"', isChecked ? 'checked' : undefined)}
            ${Utils.sprintf('disabled="%s"', isDisabled ? 'disabled' : undefined)} />
            <span></span>
            </label>`,
          this.header.formatters[j] && typeof value === 'string' ? value : '',
          this.options.cardView ? '</div>' : '</td>'
        ].join('')

        item[this.header.stateField] = value === true || (!!value_ || (value && value.checked))
      } else {
        value = typeof value === 'undefined' || value === null
          ? this.options.undefinedText : value

        if (this.options.cardView) {
          const cardTitle = this.options.showHeader
            ? `<span class="card-view-title"${style_}>${Utils.getFieldTitle(this.columns, field)}</span>` : ''

          text = `<div class="card-view">${cardTitle}<span class="card-view-value">${value}</span></div>`

          if (this.options.smartDisplay && value === '') {
            text = '<div class="card-view"></div>'
          }
        } else {
          text = `<td${id_}${class_}${style_}${data_}${rowspan_}${colspan_}${title_}>${value}</td>`
        }
      }

      html.push(text)
    })

    if (this.options.cardView) {
      html.push('</div></td>')
    }
    html.push('</tr>')

    return html.join('')
  }

  initBody (fixedScroll) {
    const data = this.getData()

    this.trigger('pre-body', data)

    this.$body = find(this.$el, 'tbody')
    if (!this.$body.length) {
      this.$body = createElem('tbody')
      appendTo(this.$el, this.$body)
    } else {
      this.$body = this.$body[0]
    }

    // Fix #389 Bootstrap-table-flatJSON is not working
    if (!this.options.pagination || this.options.sidePagination === 'server') {
      this.pageFrom = 1
      this.pageTo = data.length
    }

    const rows = []
    let hasTr = false

    for (let i = this.pageFrom - 1; i < this.pageTo; i++) {
      const item = data[i]
      const tr = this.initRow(item, i)
      hasTr = hasTr || !!tr
      if (tr && typeof tr === 'string') {
        rows.push(tr)
      }
    }

    // show no records
    if (!hasTr) {
      this.$body.appendChild(createElem(`<tr class="no-records-found">${Utils.sprintf('<td colspan="%s">%s</td>',
        this.$header.find('th').length,
        this.options.formatNoMatches())}</tr>`))
    } else {
      if (this.virtualScroll) {
        this.virtualScroll.destroy()
      }
      this.virtualScroll = new VirtualScroll({
        rows,
        scrollEl: this.$tableBody[0],
        contentEl: this.$body,
        callback: () => {
          this.fitHeader()
        }
      })
    }

    if (!fixedScroll) {
      this.scrollTo(0)
    }

    // click to select by column
    const tds = find(this.$body, 'tr[data-index] > td')
    removeEvent(tds, 'click dblclick')
    addEvent(tds, 'click dblclick', (e) => {
      const $td = $(e.currentTarget)
      const $tr = $td.parent()
      const $cardViewArr = $(e.target).parents('.card-views').children()
      const $cardViewTarget = $(e.target).parents('.card-view')
      const rowIndex = $tr.data('index')
      const item = this.data[rowIndex]
      const index = this.options.cardView ? $cardViewArr.index($cardViewTarget) : $td[0].cellIndex
      const fields = this.getVisibleFields()
      const field = fields[this.options.detailView && !this.options.cardView ? index - 1 : index]
      const column = this.columns[this.fieldsColumnsIndex[field]]
      const value = Utils.getItemField(item, field, this.options.escape)

      if ($td.find('.detail-icon').length) {
        return
      }

      this.trigger(e.type === 'click' ? 'click-cell' : 'dbl-click-cell', field, value, item, $td)
      this.trigger(e.type === 'click' ? 'click-row' : 'dbl-click-row', item, $tr, field)

      // if click to select - then trigger the checkbox/radio click
      if (
        e.type === 'click' &&
        this.options.clickToSelect &&
        column.clickToSelect &&
        !Utils.calculateObjectValue(this.options, this.options.ignoreClickToSelectOn, [e.target])
      ) {
        const $selectItem = $tr.find(Utils.sprintf('[name="%s"]', this.options.selectItemName))
        if ($selectItem.length) {
          $selectItem[0].click()
        }
      }

      if (e.type === 'click' && this.options.detailViewByClick) {
        this.toggleDetailView(rowIndex)
      }
    })

    removeEvent(tds, 'mousedown')
    addEvent(tds, 'mousedown', (e) => {
      // https://github.com/jquery/jquery/issues/1741
      this.multipleSelectRowCtrlKey = e.ctrlKey || e.metaKey
      this.multipleSelectRowShiftKey = e.shiftKey
    })

    const detailsIcons = find(this.$body, 'tr[data-index] > td > .detail-icon')
    removeEvent(detailsIcons, 'click')
    addEvent(detailsIcons, 'click', (e) => {
      e.preventDefault()
      this.toggleDetailView($(e.currentTarget).parent().parent().data('index'))
      return false
    })

    this.$selectItem = find(this.$body, Utils.sprintf('[name="%s"]', this.options.selectItemName))
    removeEvent(this.$selectItem, 'click')
    addEvent(this.$selectItem, 'click', (e) => {
      e.stopImmediatePropagation()

      const $this = $(e.currentTarget)
      this.check_($this.prop('checked'), $this.data('index'))
    })

    // Still support to jQuery
    this.$selectItem = $(this.$selectItem)

    this.header.events.forEach((_events, i) => {
      let events = _events
      if (!events) {
        return
      }
      // fix bug, if events is defined with namespace
      if (typeof events === 'string') {
        events = Utils.calculateObjectValue(null, events)
      }

      const field = this.header.fields[i]
      let fieldIndex = this.getVisibleFields().indexOf(field)

      if (fieldIndex === -1) {
        return
      }

      if (this.options.detailView && !this.options.cardView) {
        fieldIndex += 1
      }

      for (const [key, event] of Object.entries(events)) {
        find(this.$body, 'tr:not(.no-records-found)').forEach((tr, i) => {
          const $tr = tr
          const $td = find($tr, this.options.cardView ? '.card-view' : 'td')[fieldIndex]
          const index = key.indexOf(' ')
          const name = key.substring(0, index)
          const el = key.substring(index + 1)

          const tdElems = find($td, el)
          removeEvent(tdElems, name)
          addEvent(tdElems, name, (e) => {
            const index = $tr.data('index')
            const row = this.data[index]
            const value = row[field]

            event.apply(this, [e, value, row, index])
          })
        })
      }
    })

    this.updateSelected()
    this.initFooter()
    this.resetView()

    if (this.options.sidePagination !== 'server') {
      this.options.totalRows = data.length
    }

    this.trigger('post-body', data)
  }

  toggleDetailView (index) {
    const $tr = this.$body.find(Utils.sprintf('> tr[data-index="%s"]', index))

    if ($tr.next().is('tr.detail-view')) {
      this.collapseRow(index)
    } else {
      this.expandRow(index)
    }

    this.resetView()
  }

  initServer (silent, query, url) {
    let data = {}
    const index = this.header.fields.indexOf(this.options.sortName)

    let params = {
      searchText: this.searchText,
      sortName: this.options.sortName,
      sortOrder: this.options.sortOrder
    }

    if (this.header.sortNames[index]) {
      params.sortName = this.header.sortNames[index]
    }

    if (this.options.pagination && this.options.sidePagination === 'server') {
      params.pageSize = this.options.pageSize === this.options.formatAllRows()
        ? this.options.totalRows : this.options.pageSize
      params.pageNumber = this.options.pageNumber
    }

    if (!(url || this.options.url) && !this.options.ajax) {
      return
    }

    if (this.options.queryParamsType === 'limit') {
      params = {
        search: params.searchText,
        sort: params.sortName,
        order: params.sortOrder
      }

      if (this.options.pagination && this.options.sidePagination === 'server') {
        params.offset = this.options.pageSize === this.options.formatAllRows()
          ? 0 : this.options.pageSize * (this.options.pageNumber - 1)
        params.limit = this.options.pageSize === this.options.formatAllRows()
          ? this.options.totalRows : this.options.pageSize
        if (params.limit === 0) {
          delete params.limit
        }
      }
    }

    if (!(isEmptyObject(this.filterColumnsPartial))) {
      params.filter = JSON.stringify(this.filterColumnsPartial, null)
    }

    data = Utils.calculateObjectValue(this.options, this.options.queryParams, [params], data)

    Utils.extend(data, query || {})

    // false to stop request
    if (data === false) {
      return
    }

    if (!silent) {
      this.showLoading()
    }
    const request = Utils.extend({}, Utils.calculateObjectValue(null, this.options.ajaxOptions), {
      type: this.options.method,
      url: url || this.options.url,
      data: this.options.contentType === 'application/json' && this.options.method === 'post'
        ? JSON.stringify(data) : data,
      cache: this.options.cache,
      contentType: this.options.contentType,
      dataType: this.options.dataType,
      success: _res => {
        const res = Utils.calculateObjectValue(this.options,
          this.options.responseHandler, [_res], _res)

        this.load(res)
        this.trigger('load-success', res)
        if (!silent) {
          this.hideLoading()
        }
      },
      error: jqXHR => {
        let data = []
        if (this.options.sidePagination === 'server') {
          data = {}
          data[this.options.totalField] = 0
          data[this.options.dataField] = []
        }
        this.load(data)
        this.trigger('load-error', jqXHR.status, jqXHR)
        if (!silent) hide(this.$tableLoading)
      }
    })

    if (this.options.ajax) {
      Utils.calculateObjectValue(this, this.options.ajax, [request], null)
    } else {
      if (this._xhr && this._xhr.readyState !== 4) {
        this._xhr.abort()
      }
      this._xhr = $.ajax(request)
    }

    return data
  }

  initSearchText () {
    if (this.options.search) {
      this.searchText = ''
      if (this.options.searchText !== '') {
        const $search = this.$toolbar.find('.search input')
        $search.val(this.options.searchText)
        this.onSearch({currentTarget: $search, firedByInitSearchText: true})
      }
    }
  }

  getCaret () {
    find(this.$header, 'th').forEach((th, i) => {
      const sortable = find(th, '.sortable')
      removeClass(sortable, 'desc asc')
      addClass(sortable, th.getAttribute('data-field') === this.options.sortName ? this.options.sortOrder : 'both')
    })
  }

  updateSelected () {
    const checkAll = this.$selectItem.filter(':enabled').length &&
      this.$selectItem.filter(':enabled').length ===
      this.$selectItem.filter(':enabled').filter(':checked').length

    this.$selectAll.add(this.$selectAll_).prop('checked', checkAll)

    this.$selectItem.each((i, el) => {
      $(el).closest('tr')[$(el).prop('checked') ? 'addClass' : 'removeClass']('selected')
    })
  }

  updateRows () {
    this.$selectItem.each((i, el) => {
      this.data[$(el).data('index')][this.header.stateField] = $(el).prop('checked')
    })
  }

  resetRows () {
    for (const row of this.data) {
      this.$selectAll.prop('checked', false)
      this.$selectItem.prop('checked', false)
      if (this.header.stateField) {
        row[this.header.stateField] = false
      }
    }
    this.initHiddenRows()
  }

  trigger (_name, ...args) {
    const name = `${_name}.bs.table`
    this.options[BootstrapTable.EVENTS[name]](...args)
    this.$el.trigger($.Event(name), args)

    this.options.onAll(name, args)
    this.$el.trigger($.Event('all.bs.table'), [name, args])
  }

  resetHeader () {
    // fix #61: the hidden table reset header bug.
    // fix bug: get $el.css('width') error sometime (height = 500)
    clearTimeout(this.timeoutId_)
    this.timeoutId_ = setTimeout(() => this.fitHeader(), this.$el.is(':hidden') ? 100 : 0)
  }

  fitHeader () {
    if (this.$el.is(':hidden')) {
      this.timeoutId_ = setTimeout(() => this.fitHeader(), 100)
      return
    }

    const fixedBody = this.$tableBody.get(0)
    const scrollWidth = fixedBody.scrollWidth > fixedBody.clientWidth &&
    fixedBody.scrollHeight > fixedBody.clientHeight + this.$header.offsetHeight
      ? Utils.getScrollBarWidth() : 0

    this.$el.css('margin-top', -this.$header.offsetHeight)

    const focused = $(':focus')
    if (focused.length > 0) {
      const $th = focused.parents('th')
      if ($th.length > 0) {
        const dataField = $th.attr('data-field')
        if (dataField !== undefined) {
          const $headerTh = this.$header.find(`[data-field='${dataField}']`)
          if ($headerTh.length > 0) {
            addClass($headerTh.find(':input'), 'focus-temp')
          }
        }
      }
    }

    this.$header_ = this.$header.cloneNode(true)
    this.$selectAll_ = find(this.$header_, '[name="btSelectAll"]')
    this.$tableHeader
      .css('margin-right', scrollWidth)
      .find('table').css('width', this.$el.outerWidth())
      .html('').attr('class', this.$el.attr('class'))
      .append(this.$header_)

    this.$tableLoading.css('width', this.$el.outerWidth())

    const focusedTemp = $('.focus-temp:visible:eq(0)')
    if (focusedTemp.length > 0) {
      focusedTemp.focus()
      removeClass(this.$header.find('.focus-temp'), 'focus-temp')
    }

    find(this.$header, 'th[data-field]').forEach((el, i) => {
      find(this.$header_, Utils.sprintf('th[data-field="%s"]', el.getAttribute('data-field'))).forEach((el_, i_) => {
        el_ = el.cloneNode(true)
      })
    })

    const visibleFields = this.getVisibleFields()
    const $ths = find(this.$header_, 'th')
    let $tr = find(this.$body, 'tr:not(.no-records-found):not(.virtual-scroll-top)')[0]

    while ($tr && find($tr, 'td[colspan]:not([colspan="1"])').length) {
      $tr = $tr.nextSibling
    }

    find($tr, 'td').forEach((el, i) => {
      const $this = el
      let index = i

      if (this.options.detailView && !this.options.cardView) {
        if (i === 0) {
          const $thDetail = $ths.filter('.detail')
          const zoomWidth = $thDetail.offsetWidth - find($thDetail, '.fht-cell')[0].offsetWidth
          find($thDetail, '.fht-cell')[0].style.width = $this.clientWidth - zoomWidth
        }
        index = i - 1
      }

      if (index === -1) {
        return
      }

      let $th = find(this.$header_, Utils.sprintf('th[data-field="%s"]', visibleFields[index]))[0]
      if ($th.length > 1) {
        $th = $ths[$this.cellIndex]
      }

      const zoomWidth = $th.offsetWidth - find($th, '.fht-cell')[0].offsetWidth
      find($th, '.fht-cell')[0].style.width = $this.clientWidth - zoomWidth
    })

    this.horizontalScroll()
    this.trigger('post-header')
  }

  initFooter () {
    if (!this.options.showFooter || this.options.cardView) { // do nothing
      return
    }

    const data = this.getData()
    const html = []

    if (!this.options.cardView && this.options.detailView && this.options.detailViewIcon) {
      html.push('<th class="detail"><div class="th-inner"></div><div class="fht-cell"></div></th>')
    }

    for (const column of this.columns) {
      let falign = ''

      let valign = ''
      const csses = []
      let style = {}
      let class_ = Utils.sprintf(' class="%s"', column['class'])

      if (!column.visible) {
        continue
      }

      if (this.options.cardView && (!column.cardVisible)) {
        return
      }

      falign = Utils.sprintf('text-align: %s; ', column.falign ? column.falign : column.align)
      valign = Utils.sprintf('vertical-align: %s; ', column.valign)

      style = Utils.calculateObjectValue(null, this.options.footerStyle, [column])

      if (style && style.css) {
        for (const [key, value] of Object.entries(style.css)) {
          csses.push(`${key}: ${value}`)
        }
      }
      if (style && style.classes) {
        class_ = Utils.sprintf(' class="%s"', column['class'] ?
          [column['class'], style.classes].join(' ') : style.classes)
      }

      html.push('<th', class_, Utils.sprintf(' style="%s"', falign + valign + csses.concat().join('; ')), '>')
      html.push('<div class="th-inner">')

      html.push(Utils.calculateObjectValue(column, column.footerFormatter,
        [data], this.footerData[0] && this.footerData[0][column.field] || ''))

      html.push('</div>')
      html.push('<div class="fht-cell"></div>')
      html.push('</div>')
      html.push('</th>')
    }

    this.$tableFooter.find('tr').html(html.join(''))

    this.trigger('post-footer', this.$tableFooter)
  }

  fitFooter () {
    if (this.$el.is(':hidden')) {
      setTimeout(() => this.fitFooter(), 100)
      return
    }

    const fixedBody = this.$tableBody.get(0)
    const scrollWidth = fixedBody.scrollWidth > fixedBody.clientWidth &&
    fixedBody.scrollHeight > fixedBody.clientHeight + this.$header.offsetHeight
      ? Utils.getScrollBarWidth() : 0

    this.$tableFooter
      .css('margin-right', scrollWidth)
      .find('table').css('width', this.$el.outerWidth())
      .attr('class', this.$el.attr('class'))

    const visibleFields = this.getVisibleFields()
    const $ths = this.$tableFooter.find('th')
    let $tr = this.$body.find('>tr:first-child:not(.no-records-found)')

    while ($tr.length && $tr.find('>td[colspan]:not([colspan="1"])').length) {
      $tr = $tr.next()
    }

    $tr.find('> *').each((i, el) => {
      const $this = $(el)
      let index = i

      if (this.options.detailView && !this.options.cardView) {
        if (i === 0) {
          const $thDetail = $ths.filter('.detail')
          const zoomWidth = $thDetail.width() - $thDetail.find('.fht-cell').width()
          $thDetail.find('.fht-cell').width($this.innerWidth() - zoomWidth)
        }
        index = i - 1
      }

      if (index === -1) {
        return
      }

      const $th = $ths.eq(i)
      const zoomWidth = $th.width() - $th.find('.fht-cell').width()
      $th.find('.fht-cell').width($this.innerWidth() - zoomWidth)
    })

    this.horizontalScroll()
  }

  horizontalScroll () {
    // horizontal scroll event
    // TODO: it's probably better improving the layout than binding to scroll event

    this.trigger('scroll-body')
    this.$tableBody.off('scroll').on('scroll', ({currentTarget}) => {
      if (this.options.showHeader && this.options.height) {
        this.$tableHeader.scrollLeft($(currentTarget).scrollLeft())
      }

      if (this.options.showFooter && !this.options.cardView) {
        this.$tableFooter.scrollLeft($(currentTarget).scrollLeft())
      }
    })
  }

  toggleColumn (index, checked, needUpdate) {
    if (index === -1) {
      return
    }
    this.columns[index].visible = checked
    this.initHeader()
    this.initSearch()
    this.initPagination()
    this.initBody()

    if (this.options.showColumns) {
      const $items = this.$toolbar.find('.keep-open input').prop('disabled', false)

      if (needUpdate) {
        $items.filter(Utils.sprintf('[value="%s"]', index)).prop('checked', checked)
      }

      if ($items.filter(':checked').length <= this.options.minimumCountColumns) {
        $items.filter(':checked').prop('disabled', true)
      }
    }
  }

  getVisibleFields () {
    const visibleFields = []

    for (const field of this.header.fields) {
      const column = this.columns[this.fieldsColumnsIndex[field]]

      if (!column.visible) {
        continue
      }
      visibleFields.push(field)
    }
    return visibleFields
  }

  // PUBLIC FUNCTION DEFINITION
  // =======================

  resetView (params) {
    let padding = 0

    if (params && params.height) {
      this.options.height = params.height
    }

    this.$selectAll.prop('checked', this.$selectItem.length > 0 &&
      this.$selectItem.length === this.$selectItem.filter(':checked').length)

    if (this.options.cardView) {
      addClass(this.$tableContainer, 'has-card-view')
    } else {
      removeClass(this.$tableContainer, 'has-card-view')
    }

    if (!this.options.cardView && this.options.showHeader && this.options.height) {
      show(this.$tableHeader)
      this.resetHeader()
      padding += outerHeight(this.$header)
    } else {
      hide(this.$tableHeader)
      this.trigger('post-header')
    }

    if (!this.options.cardView && this.options.showFooter) {
      show(this.$tableFooter)
      this.fitFooter()
      if (this.options.height) {
        padding += outerHeight(this.$tableFooter)
      }
    }

    if (this.options.height) {
      const toolbarHeight = outerHeight(this.$toolbar)
      const paginationHeight = outerHeight(this.$pagination)
      const height = this.options.height - toolbarHeight - paginationHeight
      const tableHeight = outerHeight(this.$tableBody.find('table'))
      this.$tableContainer[0].style.height = `${height}px`
      this.$tableBorder && this.$tableBorder.css('height', `${height - tableHeight - padding - 1}px`)
    }

    if (this.options.cardView) {
      // remove the element css
      this.$el.css('margin-top', '0')
      this.$tableContainer[0].style.paddingBottom = '0'
      hide(this.$tableFooter)
    } else {
      // Assign the correct sortable arrow
      this.getCaret()
      this.$tableContainer[0].style.paddingBottom = `${padding}px`
    }

    this.trigger('reset-view')
  }

  getData (useCurrentPage) {
    let data = this.options.data
    if (this.searchText || this.options.sortName || !isEmptyObject(this.filterColumns) || !isEmptyObject(this.filterColumnsPartial)) {
      data = this.data
    }

    if (useCurrentPage) {
      return data.slice(this.pageFrom - 1, this.pageTo)
    }

    return data
  }

  load (_data) {
    let fixedScroll = false
    let data = _data

    // #431: support pagination
    if (this.options.pagination && this.options.sidePagination === 'server') {
      this.options.totalRows = data[this.options.totalField]
    }

    if (this.options.pagination && this.options.sidePagination === 'server') {
      this.options.totalNotFiltered = data[this.options.totalNotFilteredField]
    }

    fixedScroll = data.fixedScroll
    data = Array.isArray(data) ? data : data[this.options.dataField]

    this.initData(data)
    this.initSearch()
    this.initPagination()
    this.initBody(fixedScroll)
  }

  append (data) {
    this.initData(data, 'append')
    this.initSearch()
    this.initPagination()
    this.initSort()
    this.initBody(true)
  }

  prepend (data) {
    this.initData(data, 'prepend')
    this.initSearch()
    this.initPagination()
    this.initSort()
    this.initBody(true)
  }

  remove (params) {
    const len = this.options.data.length
    let i
    let row

    if (!params.hasOwnProperty('field') || !params.hasOwnProperty('values')) {
      return
    }

    for (i = len - 1; i >= 0; i--) {
      row = this.options.data[i]

      if (!row.hasOwnProperty(params.field)) {
        continue
      }
      if (params.values.includes(row[params.field])) {
        this.options.data.splice(i, 1)
        if (this.options.sidePagination === 'server') {
          this.options.totalRows -= 1
        }
      }
    }

    if (len === this.options.data.length) {
      return
    }

    this.initSearch()
    this.initPagination()
    this.initSort()
    this.initBody(true)
  }

  removeAll () {
    if (this.options.data.length > 0) {
      this.options.data.splice(0, this.options.data.length)
      this.initSearch()
      this.initPagination()
      this.initBody(true)
    }
  }

  getRowByUniqueId (_id) {
    const uniqueId = this.options.uniqueId
    const len = this.options.data.length
    let id = _id
    let dataRow = null
    let i
    let row
    let rowUniqueId

    for (i = len - 1; i >= 0; i--) {
      row = this.options.data[i]

      if (row.hasOwnProperty(uniqueId)) { // uniqueId is a column
        rowUniqueId = row[uniqueId]
      } else if (row._data && row._data.hasOwnProperty(uniqueId)) { // uniqueId is a row data property
        rowUniqueId = row._data[uniqueId]
      } else {
        continue
      }

      if (typeof rowUniqueId === 'string') {
        id = id.toString()
      } else if (typeof rowUniqueId === 'number') {
        if ((Number(rowUniqueId) === rowUniqueId) && (rowUniqueId % 1 === 0)) {
          id = parseInt(id)
        } else if ((rowUniqueId === Number(rowUniqueId)) && (rowUniqueId !== 0)) {
          id = parseFloat(id)
        }
      }

      if (rowUniqueId === id) {
        dataRow = row
        break
      }
    }

    return dataRow
  }

  removeByUniqueId (id) {
    const len = this.options.data.length
    const row = this.getRowByUniqueId(id)

    if (row) {
      this.options.data.splice(this.options.data.indexOf(row), 1)
    }

    if (len === this.options.data.length) {
      return
    }

    this.initSearch()
    this.initPagination()
    this.initBody(true)
  }

  updateByUniqueId (params) {
    const allParams = Array.isArray(params) ? params : [params]

    for (const params of allParams) {
      if (!params.hasOwnProperty('id') || !params.hasOwnProperty('row')) {
        continue
      }

      const rowId = this.options.data.indexOf(this.getRowByUniqueId(params.id))

      if (rowId === -1) {
        continue
      }
      Utils.extend(this.options.data[rowId], params.row)
    }

    this.initSearch()
    this.initPagination()
    this.initSort()
    this.initBody(true)
  }

  refreshColumnTitle (params) {
    if (!params.hasOwnProperty('field') || !params.hasOwnProperty('title')) {
      return
    }

    this.columns[this.fieldsColumnsIndex[params.field]].title =
      this.options.escape ? Utils.escapeHTML(params.title) : params.title

    if (this.columns[this.fieldsColumnsIndex[params.field]].visible) {
      const header = this.options.height !== undefined ? this.$tableHeader : this.$header
      header.find('th[data-field]').each((i, el) => {
        if ($(el).data('field') === params.field) {
          $($(el).find('.th-inner')[0]).text(params.title)
          return false
        }
      })
    }
  }

  insertRow (params) {
    if (!params.hasOwnProperty('index') || !params.hasOwnProperty('row')) {
      return
    }
    this.options.data.splice(params.index, 0, params.row)
    this.initSearch()
    this.initPagination()
    this.initSort()
    this.initBody(true)
  }

  updateRow (params) {
    const allParams = Array.isArray(params) ? params : [params]

    for (const params of allParams) {
      if (!params.hasOwnProperty('index') || !params.hasOwnProperty('row')) {
        continue
      }
      Utils.extend(this.options.data[params.index], params.row)
    }

    this.initSearch()
    this.initPagination()
    this.initSort()
    this.initBody(true)
  }

  initHiddenRows () {
    this.hiddenRows = []
  }

  showRow (params) {
    this.toggleRow(params, true)
  }

  hideRow (params) {
    this.toggleRow(params, false)
  }

  toggleRow (params, visible) {
    let row

    if (params.hasOwnProperty('index')) {
      row = this.getData()[params.index]
    } else if (params.hasOwnProperty('uniqueId')) {
      row = this.getRowByUniqueId(params.uniqueId)
    }

    if (!row) {
      return
    }

    const index = Utils.findIndex(this.hiddenRows, row)

    if (!visible && index === -1) {
      this.hiddenRows.push(row)
    } else if (visible && index > -1) {
      this.hiddenRows.splice(index, 1)
    }
    this.initBody(true)
  }

  getHiddenRows (show) {
    if (show) {
      this.initHiddenRows()
      this.initBody(true)
      return
    }
    const data = this.getData()
    const rows = []

    for (const row of data) {
      if (this.hiddenRows.includes(row)) {
        rows.push(row)
      }
    }
    this.hiddenRows = rows
    return rows
  }

  mergeCells (options) {
    const row = options.index
    let col = this.getVisibleFields().indexOf(options.field)
    const rowspan = options.rowspan || 1
    const colspan = options.colspan || 1
    let i
    let j
    const $tr = this.$body.find('>tr')

    if (this.options.detailView && !this.options.cardView) {
      col += 1
    }

    const $td = $tr.eq(row).find('>td').eq(col)

    if (row < 0 || col < 0 || row >= this.data.length) {
      return
    }

    for (i = row; i < row + rowspan; i++) {
      for (j = col; j < col + colspan; j++) {
        hide($tr.eq(i).find('>td').eq(j))
      }
    }

    show($td.attr('rowspan', rowspan).attr('colspan', colspan))
  }

  updateCell (params) {
    if (!params.hasOwnProperty('index') ||
      !params.hasOwnProperty('field') ||
      !params.hasOwnProperty('value')) {
      return
    }
    this.data[params.index][params.field] = params.value

    if (params.reinit === false) {
      return
    }
    this.initSort()
    this.initBody(true)
  }

  updateCellById (params) {
    if (!params.hasOwnProperty('id') ||
      !params.hasOwnProperty('field') ||
      !params.hasOwnProperty('value')) {
      return
    }
    const allParams = Array.isArray(params) ? params : [params]

    allParams.forEach(({id, field, value}) => {
      const rowId = this.options.data.indexOf(this.getRowByUniqueId(id))

      if (rowId === -1) {
        return
      }
      this.data[rowId][field] = value
    })

    if (params.reinit === false) {
      return
    }
    this.initSort()
    this.initBody(true)
  }

  getOptions () {
    // deep copy and remove data
    const options = JSON.parse(JSON.stringify(this.options))
    delete options.data
    return options
  }

  getSelections () {
    // fix #2424: from html with checkbox
    return this.options.data.filter(row =>
      row[this.header.stateField] === true)
  }

  getAllSelections () {
    return this.options.data.filter(row => row[this.header.stateField])
  }

  checkAll () {
    this.checkAll_(true)
  }

  uncheckAll () {
    this.checkAll_(false)
  }

  checkInvert () {
    const $items = this.$selectItem.filter(':enabled')
    let checked = $items.filter(':checked')
    $items.each((i, el) => {
      $(el).prop('checked', !$(el).prop('checked'))
    })
    this.updateRows()
    this.updateSelected()
    this.trigger('uncheck-some', checked)
    checked = this.getSelections()
    this.trigger('check-some', checked)
  }

  checkAll_ (checked) {
    const rowsBefore = this.getSelections()
    this.$selectAll.add(this.$selectAll_).prop('checked', checked)
    this.$selectItem.filter(':enabled').prop('checked', checked)
    this.updateRows()

    const rowsAfter = this.getSelections()
    if (checked) {
      this.trigger('check-all', rowsAfter, rowsBefore)
      return
    }

    this.trigger('uncheck-all', rowsAfter, rowsBefore)
  }

  check (index) {
    this.check_(true, index)
  }

  uncheck (index) {
    this.check_(false, index)
  }

  check_ (checked, index) {
    const $el = this.$selectItem.filter(`[data-index="${index}"]`)
    const row = this.data[index]

    if (
      $el.is(':radio') ||
      this.options.singleSelect ||
      this.options.multipleSelectRow &&
      !this.multipleSelectRowCtrlKey &&
      !this.multipleSelectRowShiftKey
    ) {
      for (const r of this.options.data) {
        r[this.header.stateField] = false
      }
      this.$selectItem.filter(':checked').not($el).prop('checked', false)
    }

    row[this.header.stateField] = checked

    if (this.options.multipleSelectRow) {
      if (this.multipleSelectRowShiftKey && this.multipleSelectRowLastSelectedIndex >= 0) {
        const indexes = [this.multipleSelectRowLastSelectedIndex, index].sort()

        for (let i = indexes[0] + 1; i < indexes[1]; i++) {
          this.data[i][this.header.stateField] = true
          this.$selectItem.filter(`[data-index="${i}"]`).prop('checked', true)
        }
      }

      this.multipleSelectRowCtrlKey = false
      this.multipleSelectRowShiftKey = false
      this.multipleSelectRowLastSelectedIndex = checked ? index : -1
    }

    $el.prop('checked', checked)
    this.updateSelected()
    this.trigger(checked ? 'check' : 'uncheck', this.data[index], $el)
  }

  checkBy (obj) {
    this.checkBy_(true, obj)
  }

  uncheckBy (obj) {
    this.checkBy_(false, obj)
  }

  checkBy_ (checked, obj) {
    if (!obj.hasOwnProperty('field') || !obj.hasOwnProperty('values')) {
      return
    }

    const rows = []
    this.options.data.forEach((row, i) => {
      if (!row.hasOwnProperty(obj.field)) {
        return false
      }
      if (obj.values.includes(row[obj.field])) {
        const $el = this.$selectItem.filter(':enabled')
          .filter(Utils.sprintf('[data-index="%s"]', i)).prop('checked', checked)
        row[this.header.stateField] = checked
        rows.push(row)
        this.trigger(checked ? 'check' : 'uncheck', row, $el)
      }
    })
    this.updateSelected()
    this.trigger(checked ? 'check-some' : 'uncheck-some', rows)
  }

  destroy () {
    this.$el.insertBefore(this.$container)
    $(this.options.toolbar).insertBefore(this.$el)
    this.$container.next().remove()
    this.$container.remove()
    this.$el.html(this.$el_.html())
      .css('margin-top', '0')
      .attr('class', this.$el_.attr('class') || '') // reset the class
  }

  showLoading () {
    this.$tableLoading.css('display', 'flex')
  }

  hideLoading () {
    this.$tableLoading.css('display', 'none')
  }

  togglePagination () {
    this.options.pagination = !this.options.pagination
    this.$toolbar.find('button[name="paginationSwitch"]')
      .html(Utils.sprintf(this.constants.html.icon, this.options.iconsPrefix,
        this.options.pagination ? this.options.icons.paginationSwitchDown : this.options.icons.paginationSwitchUp))
    this.updatePagination()
  }

  toggleFullscreen () {
    this.$el.closest('.bootstrap-table').toggleClass('fullscreen')
    this.resetView()
  }

  refresh (params) {
    if (params && params.url) {
      this.options.url = params.url
    }
    if (params && params.pageNumber) {
      this.options.pageNumber = params.pageNumber
    }
    if (params && params.pageSize) {
      this.options.pageSize = params.pageSize
    }
    this.trigger('refresh', this.initServer(params && params.silent,
      params && params.query, params && params.url))
  }

  resetWidth () {
    if (this.options.showHeader && this.options.height) {
      this.fitHeader()
    }
    if (this.options.showFooter && !this.options.cardView) {
      this.fitFooter()
    }
  }

  showColumn (field) {
    const fields = Array.isArray(field) ? field : [field]
    fields.forEach(field => {
      this.toggleColumn(this.fieldsColumnsIndex[field], true, true)
    })
  }

  hideColumn (field) {
    const fields = Array.isArray(field) ? field : [field]
    fields.forEach(field => {
      this.toggleColumn(this.fieldsColumnsIndex[field], false, true)
    })
  }

  getHiddenColumns () {
    return this.columns.filter(({visible}) => !visible)
  }

  getVisibleColumns () {
    return this.columns.filter(({visible}) => visible)
  }

  toggleAllColumns (visible) {
    for (const column of this.columns) {
      column.visible = visible
    }

    this.initHeader()
    this.initSearch()
    this.initPagination()
    this.initBody()
    if (this.options.showColumns) {
      const $items = this.$toolbar.find('.keep-open input').prop('disabled', false)

      if ($items.filter(':checked').length <= this.options.minimumCountColumns) {
        $items.filter(':checked').prop('disabled', true)
      }
    }
  }

  showAllColumns () {
    this.toggleAllColumns(true)
  }

  hideAllColumns () {
    this.toggleAllColumns(false)
  }

  filterBy (columns, options) {
    this.filterOptions = isEmptyObject(options) ? this.options.filterOptions : Utils.extend(this.options.filterOptions, options)
    this.filterColumns = isEmptyObject(columns) ? {} : columns
    this.options.pageNumber = 1
    this.initSearch()
    this.updatePagination()
  }

  scrollTo (params) {
    if (typeof params === 'undefined') {
      return this.$tableBody.scrollTop()
    }

    let options = {unit: 'px', value: 0}
    if (typeof params === 'object') {
      options = Object.assign(options, params)
    } else if (typeof params === 'string' && params === 'bottom') {
      options.value = this.$tableBody[0].scrollHeight
    } else if (typeof params === 'string') {
      options.value = params
    }

    let scrollTo = options.value
    if (options.unit === 'rows') {
      scrollTo = 0
      this.$body.find(`> tr:lt(${options.value})`).each((i, el) => {
        scrollTo += outerHeight($(el))
      })
    }

    this.$tableBody.scrollTop(scrollTo)
  }

  getScrollPosition () {
    return this.scrollTo()
  }

  selectPage (page) {
    if (page > 0 && page <= this.options.totalPages) {
      this.options.pageNumber = page
      this.updatePagination()
    }
  }

  prevPage () {
    if (this.options.pageNumber > 1) {
      this.options.pageNumber--
      this.updatePagination()
    }
  }

  nextPage () {
    if (this.options.pageNumber < this.options.totalPages) {
      this.options.pageNumber++
      this.updatePagination()
    }
  }

  toggleView () {
    this.options.cardView = !this.options.cardView
    this.initHeader()
    // Fixed remove toolbar when click cardView button.
    // this.initToolbar();
    this.$toolbar.find('button[name="toggle"]')
      .html(Utils.sprintf(this.constants.html.icon, this.options.iconsPrefix,
        this.options.cardView ? this.options.icons.toggleOn : this.options.icons.toggleOff))
    this.initBody()
    this.trigger('toggle', this.options.cardView)
  }

  refreshOptions (options) {
    // If the objects are equivalent then avoid the call of destroy / init methods
    if (Utils.compareObjects(this.options, options, true)) {
      return
    }
    this.options = Utils.extend(this.options, options)
    this.trigger('refresh-options', this.options)
    this.destroy()
    this.init()
  }

  resetSearch (text) {
    const $search = this.$toolbar.find('.search input')
    $search.val(text || '')
    this.onSearch({currentTarget: $search})
  }

  expandRow (index) {
    const row = this.data[index]
    const $tr = this.$body.find(Utils.sprintf('> tr[data-index="%s"][data-has-detail-view]', index))
    if ($tr.next().is('tr.detail-view')) {
      return
    }

    if (this.options.detailViewIcon) {
      $tr.find('a.detail-icon').html(Utils.sprintf(this.constants.html.icon, this.options.iconsPrefix, this.options.icons.detailClose))
    }

    $tr.after(Utils.sprintf('<tr class="detail-view"><td colspan="%s"></td></tr>', $tr.children('td').length))

    const $element = $tr.next().find('td')

    const detailFormatter = this.header.detailFormatters[index - 1] || this.options.detailFormatter
    const content = Utils.calculateObjectValue(this.options, detailFormatter, [index, row, $element], '')
    if ($element.length === 1) {
      $element.append(content)
    }

    this.trigger('expand-row', index, row, $element)
  }

  collapseRow (index) {
    const row = this.data[index]
    const $tr = this.$body.find(Utils.sprintf('> tr[data-index="%s"][data-has-detail-view]', index))
    if (!$tr.next().is('tr.detail-view')) {
      return
    }

    if (this.options.detailViewIcon) {
      $tr.find('a.detail-icon').html(Utils.sprintf(this.constants.html.icon, this.options.iconsPrefix, this.options.icons.detailOpen))
    }

    this.trigger('collapse-row', index, row, $tr.next())
    $tr.next().remove()
  }

  expandAllRows () {
    const trs = this.$body.find('> tr[data-index][data-has-detail-view]')
    for (let i = 0; i < trs.length; i++) {
      this.expandRow($(trs[i]).data('index'))
    }
  }

  collapseAllRows () {
    const trs = this.$body.find('> tr[data-index][data-has-detail-view]')
    for (let i = 0; i < trs.length; i++) {
      this.collapseRow($(trs[i]).data('index'))
    }
  }

  updateFormatText (name, text) {
    if (this.options[Utils.sprintf('format%s', name)]) {
      if (typeof text === 'string') {
        this.options[Utils.sprintf('format%s', name)] = () => text
      } else if (typeof text === 'function') {
        this.options[Utils.sprintf('format%s', name)] = text
      }
    }
    this.initToolbar()
    this.initPagination()
    this.initBody()
  }
}

BootstrapTable.DEFAULTS = Constants.DEFAULTS
BootstrapTable.LOCALES = Constants.LOCALES
BootstrapTable.COLUMN_DEFAULTS = Constants.COLUMN_DEFAULTS
BootstrapTable.EVENTS = Constants.EVENTS

// BOOTSTRAP TABLE PLUGIN DEFINITION
// =======================

const allowedMethods = [
  'getOptions',
  'getSelections', 'getAllSelections', 'getData',
  'load', 'append', 'prepend', 'remove', 'removeAll',
  'insertRow', 'updateRow', 'updateCell',
  'updateByUniqueId', 'removeByUniqueId',
  'getRowByUniqueId', 'showRow', 'hideRow', 'getHiddenRows',
  'mergeCells', 'refreshColumnTitle',
  'checkAll', 'uncheckAll', 'checkInvert',
  'check', 'uncheck',
  'checkBy', 'uncheckBy',
  'refresh',
  'resetView',
  'resetWidth',
  'destroy',
  'showLoading', 'hideLoading',
  'showColumn', 'hideColumn',
  'getHiddenColumns', 'getVisibleColumns',
  'showAllColumns', 'hideAllColumns',
  'filterBy',
  'scrollTo',
  'getScrollPosition',
  'selectPage', 'prevPage', 'nextPage',
  'togglePagination',
  'toggleView',
  'refreshOptions',
  'resetSearch',
  'expandRow', 'collapseRow', 'toggleDetailView',
  'expandAllRows', 'collapseAllRows',
  'updateFormatText', 'updateCellById'
]

$.BootstrapTable = BootstrapTable
$.fn.bootstrapTable = function (option, ...args) {
  let value

  this.each((i, el) => {
    let data = $(el).data('bootstrap.table')
    const options = Utils.extend({}, BootstrapTable.DEFAULTS, $(el).data(),
      typeof option === 'object' && option)

    if (typeof option === 'string') {
      if (!allowedMethods.includes(option)) {
        throw new Error(`Unknown method: ${option}`)
      }

      if (!data) {
        return
      }

      value = data[option](...args)

      if (option === 'destroy') {
        $(el).removeData('bootstrap.table')
      }
    }

    if (!data) {
      $(el).data('bootstrap.table', (data = new $.BootstrapTable(el, options)))
    }
  })

  return typeof value === 'undefined' ? this : value
}

$.fn.bootstrapTable.Constructor = BootstrapTable
$.fn.bootstrapTable.theme = Constants.THEME
$.fn.bootstrapTable.defaults = BootstrapTable.DEFAULTS
$.fn.bootstrapTable.columnDefaults = BootstrapTable.COLUMN_DEFAULTS
$.fn.bootstrapTable.locales = BootstrapTable.LOCALES
$.fn.bootstrapTable.methods = allowedMethods
$.fn.bootstrapTable.utils = Utils

// BOOTSTRAP TABLE INIT
// =======================

$(() => {
  $('[data-toggle="table"]').bootstrapTable()
})

export default BootstrapTable
